@@ -1,5 +1,5 @@
 /*
- *	$Id: ipfwadm.c,v 1.11 1996/05/05 17:08:50 jos Exp $
+ *	$Id: ipfwadm.c,v 1.13 1996/05/24 13:36:55 jos Exp $
  *
  *
  *	ipfwadm -- IP firewall and accounting administration
@@ -63,6 +63,19 @@
  *	1.10	Change layout of listing of active masqueraded sessions.
  *		Port to AXP (64-bit issues), untested.
  *	1.11	Fix bugs introduced in 1.10.
+ *		[ipfwadm 2.0]
+ *	1.12	Add support for port redirection (for supporting transparent
+ *		proxying) via the -r flag.
+ *		Support input/output only accounting rules by recognizing
+ *		a direction (in/out/both) after the -A flag.
+ *		Add -m option to use masquerading with policy accept
+ *		(as a replacement of the masquerade policy, which is still
+ *		supported for backwards compatibility).
+ *		Slightly change the layout of the rule listings.
+ *		Remove check for TCP protocol when using -k or -y.
+ *	1.13	Also support kernels 1.3.66 up to 1.99.6 (untested).
+ *		Check on missing command.
+ *		[ipfwadm 2.1]
  */
 
 #include <stdio.h>
@@ -79,6 +92,17 @@
 #include <linux/if.h>
 #include <linux/ip_fw.h>
 
+#ifndef	IP_FW_F_REDIR
+#define IPFWADM_NO_REDIR
+#define IP_FW_F_REDIR	0
+#endif	/* ! IP_FW_F_REDIR */
+
+#if	!defined(IP_FW_F_ACCTIN) || !defined(IP_FW_F_ACCTOUT)
+#define IPFWADM_NO_ACCTDIR
+#define IP_FW_F_ACCTIN	0
+#define IP_FW_F_ACCTOUT	0
+#endif	/* ! IP_FW_F_REDIR */
+
 #define IP_VERSION	4
 #define IP_OFFSET	0x1FFF
 
@@ -100,22 +124,24 @@
 #define CMD_HELP	0x0080
 #define CMD_INSERT	0x0100
 
-#define OPT_NONE	0x0000
-#define OPT_EXTENDED	0x0001
-#define OPT_NUMERIC	0x0002
-#define OPT_SOURCE	0x0004
-#define OPT_DESTINATION	0x0008
-#define OPT_PROTOCOL	0x0010
-#define OPT_VIAADDR	0x0020
-#define OPT_POLICY	0x0040
-#define OPT_TCPSYN	0x0080
-#define OPT_BIDIR	0x0100
-#define OPT_VERBOSE	0x0200
-#define OPT_PRINTK	0x0400
-#define OPT_EXPANDED	0x0800
-#define OPT_TCPACK	0x1000
-#define OPT_TOS		0x2000
-#define OPT_VIANAME	0x4000
+#define OPT_NONE	0x000000
+#define OPT_EXTENDED	0x000001
+#define OPT_NUMERIC	0x000002
+#define OPT_SOURCE	0x000004
+#define OPT_DESTINATION	0x000008
+#define OPT_PROTOCOL	0x000010
+#define OPT_VIAADDR	0x000020
+#define OPT_POLICY	0x000040
+#define OPT_TCPSYN	0x000080
+#define OPT_BIDIR	0x000100
+#define OPT_VERBOSE	0x000200
+#define OPT_PRINTK	0x000400
+#define OPT_EXPANDED	0x000800
+#define OPT_TCPACK	0x001000
+#define OPT_TOS		0x002000
+#define OPT_VIANAME	0x004000
+#define OPT_REDIRECT	0x008000
+#define OPT_MASQUERADE	0x010000
 
 #define FMT_NUMERIC	0x0001
 #define FMT_NOCOUNTS	0x0002
@@ -140,8 +166,8 @@ struct masq {
 	short		pdelta;		/* Delta in sequence numbers before last */
 };
 
-char ipfwadm_version[] = "$Id: ipfwadm.c,v 1.11 1996/05/05 17:08:50 jos Exp $";
-char package_version[] = "ipfwadm 2.0, 1996/05/05";
+char ipfwadm_version[] = "$Id: ipfwadm.c,v 1.13 1996/05/24 13:36:55 jos Exp $";
+char package_version[] = "ipfwadm 2.1, 1996/05/24";
 
 int ssocmd_insert[4] =
 	{IP_FW_INSERT_FWD, IP_FW_INSERT_IN, IP_FW_INSERT_OUT, IP_ACCT_INSERT};
@@ -162,7 +188,7 @@ char *procfiles[5] = {"/proc/net/ip_forward", "/proc/net/ip_input",
 
 int chain = CHN_NONE;
 int command = CMD_NONE;
-int options = OPT_NONE;
+long options = OPT_NONE;
 
 char *program;
 
@@ -170,6 +196,7 @@ struct ip_fw firewall;
 
 char *sports[IP_FW_MAX_PORTS];
 char *dports[IP_FW_MAX_PORTS];
+char *rport;
 char *shostnetworkmask, *dhostnetworkmask;
 int nsaddrs, ndaddrs;
 struct in_addr *saddrs, *daddrs;
@@ -179,6 +206,7 @@ extern void parse_hostnetworkmask(char *, struct in_addr **,
 	struct in_addr *, int *);
 extern void parse_viahost(char *);
 extern struct in_addr *parse_mask(char *);
+extern void parse_direction(char *);
 extern void parse_policy(char *);
 extern void parse_protocol(char *);
 extern void parse_all_ports(char **, unsigned short *, int, int);
@@ -211,7 +239,7 @@ extern int read_masqinfo(FILE *, struct masq *, int);
 extern struct ip_fwpkt *fw_to_fwpkt(struct ip_fw *);
 extern int do_setsockopt(int, void *, int);
 
-extern void check_option(int, char);
+extern void check_option(long, char);
 extern void inaddrcpy(struct in_addr *, struct in_addr *);
 extern void *fw_malloc(size_t);
 extern void *fw_calloc(size_t, size_t);
@@ -227,12 +255,15 @@ main(int argc, char *argv[])
 
 	program = argv[0];
 
-	while ((c = getopt(argc, argv, "AFIMOadilzfp:chP:S:D:V:W:beknot:vxy")) != -1)
+	while ((c = getopt(argc, argv, "AFIMOadilzfp:chP:S:D:V:W:bekmnor:t:vxy")) != -1)
 		switch (c) {
 		case 'A':
 			if (chain != CHN_NONE)
 				exit_error(2, "multiple categories specified");
 			chain = CHN_ACCT;
+			if (optind < argc && argv[optind][0] != '-') {
+				parse_direction(argv[optind++]);
+			}
 			break;
 		case 'F':
 			if (chain != CHN_NONE)
@@ -348,6 +379,10 @@ main(int argc, char *argv[])
 			check_option(OPT_TCPACK, 'k');
 			firewall.fw_flg |= IP_FW_F_TCPACK;
 			break;
+		case 'm':
+			check_option(OPT_MASQUERADE, 'm');
+			firewall.fw_flg |= IP_FW_F_MASQ;
+			break;
 		case 'n':
 			check_option(OPT_NUMERIC, 'n');
 			break;
@@ -355,6 +390,15 @@ main(int argc, char *argv[])
 			check_option(OPT_PRINTK, 'o');
 			firewall.fw_flg |= IP_FW_F_PRN;
 			break;
+		case 'r':
+#ifndef	IPFWADM_NO_REDIR
+			check_option(OPT_REDIRECT, 'r');
+			firewall.fw_flg |= IP_FW_F_REDIR;
+			rport = optarg;
+#else	/* IPFWADM_NO_REDIR */
+			exit_error(2, "redirection not supported");
+#endif	/* IPFWADM_NO_REDIR */
+			break;
 		case 't':
 			check_option(OPT_TOS, 't');
 			if (optind < argc && argv[optind][0] != '-') {
@@ -380,6 +424,8 @@ main(int argc, char *argv[])
 
 	if (optind < argc)
 		exit_error(2, "unknown arguments found on commandline");
+	else if (!command)
+		exit_error(2, "no command specified");
 
 	kind = firewall.fw_flg & IP_FW_F_KIND;
 	policy = (int) (firewall.fw_flg &
@@ -410,6 +456,9 @@ main(int argc, char *argv[])
 	else if ((options & OPT_PRINTK) &&
 			!(command & (CMD_INSERT | CMD_APPEND | CMD_DELETE)))
 		exit_error(2, "kernel print flag (-o) only allowed with insert/append/delete");
+	else if ((options & OPT_REDIRECT) &&
+			!(command & (CMD_INSERT | CMD_APPEND | CMD_DELETE)))
+		exit_error(2, "port redirection (-r) only allowed with insert/append/delete");
 	else if ((options & OPT_TOS) &&
 			!(command & (CMD_INSERT | CMD_APPEND | CMD_DELETE)))
 		exit_error(2, "tos values (-t) only allowed with insert/append/delete");
@@ -425,11 +474,6 @@ main(int argc, char *argv[])
 	if (command == CMD_CHECK && kind == IP_FW_F_ALL)
 		exit_error(2, "specific protocol required for check command");
 
-	if (firewall.fw_flg & IP_FW_F_TCPSYN && kind != IP_FW_F_TCP)
-		exit_error(2, "SYN option (-y) only allowed for TCP protocol");
-	else if (firewall.fw_flg & IP_FW_F_TCPACK && kind != IP_FW_F_TCP)
-		exit_error(2, "ACK option (-k) only allowed for TCP protocol");
-
 	if ((options & OPT_VIAADDR) && (command & (CMD_LIST | CMD_FLUSH |
 			CMD_RESET | CMD_POLICY)))
 		exit_error(2, "no interface (-V) allowed with this command");
@@ -446,10 +490,16 @@ main(int argc, char *argv[])
 	else if (command == CMD_CHECK && kind == IP_FW_F_ICMP &&
 			(firewall.fw_nsp != 1 || firewall.fw_ndp != 0))
 		exit_error(2, "ICMP type required after source address");
-	else if ((policy & IP_FW_F_MASQ) && chain != CHN_FWD)
+	else if (options & OPT_REDIRECT && chain != CHN_IN)
+		exit_error(2, "redirecting only allowed in combination with -I");
+	else if (kind == IP_FW_F_ICMP && options & OPT_REDIRECT)
+		exit_error(2, "redirecting not allowed with protocol ICMP");
+	else if (options & OPT_MASQUERADE && chain != CHN_FWD)
 		exit_error(2, "masquerading only allowed in combination with -F");
-	else if (kind == IP_FW_F_ICMP && (policy & IP_FW_F_MASQ))
+	else if (kind == IP_FW_F_ICMP && options & OPT_MASQUERADE)
 		exit_error(2, "masquerading not allowed with protocol ICMP");
+	else if (options & OPT_MASQUERADE && !(policy & IP_FW_F_ACCEPT))
+		exit_error(2, "masquerading only allowed with policy accept");
 
 	if ((options & OPT_SOURCE) && (command & (CMD_LIST | CMD_FLUSH |
 			CMD_RESET | CMD_POLICY)))
@@ -477,6 +527,16 @@ main(int argc, char *argv[])
 			(int) firewall.fw_ndp, firewall.fw_flg & IP_FW_F_DRNG);
 	}
 
+	if (options & OPT_REDIRECT) {
+		if (firewall.fw_nsp + firewall.fw_ndp >= IP_FW_MAX_PORTS) {
+			fprintf(stderr, "%s: too many ports specified (maximum %d when using -r option)\n",
+				program, IP_FW_MAX_PORTS - 1);
+			exit_tryhelp(2);
+		}
+		firewall.fw_pts[firewall.fw_nsp + firewall.fw_ndp] =
+			parse_port(rport, kind);
+	}
+
 	if (!(options & OPT_TOS)) {
 		firewall.fw_tosand = 0xFF;
 		firewall.fw_tosxor = 0x00;
@@ -550,7 +610,9 @@ parse_hostnetwork(char *name, int *naddrs)
 	} else {
 		fprintf(stderr, "%s: host/network \"%s\" not found\n",
 			program, name);
-		exit(2);
+		exit_tryhelp(2);
+		/* make the compiler happy... */
+		return NULL;
 	}
 }
 
@@ -599,14 +661,14 @@ parse_viahost(char *name)
 			fprintf(stderr,
 				"%s: hostname \"%s\" does not specify a unique address\n",
 				program, name);
-			exit(2);
+			exit_tryhelp(2);
 		} else {
 			inaddrcpy(&firewall.fw_via, addrp);
 			return;
 		}
 	} else {
 		fprintf(stderr, "%s: host \"%s\" not found\n", program, name);
-		exit(2);
+		exit_tryhelp(2);
 	}
 }
 
@@ -626,7 +688,9 @@ parse_mask(char *mask)
 		return addrp;
 	} else if ((bits = string_to_number(mask, 0, 32)) == -1) {
 		fprintf(stderr, "%s: invalid mask \"%s\" specified\n", program, mask);
-		exit(2);
+		exit_tryhelp(2);
+		/* make the compiler happy... */
+		return NULL;
 	} else if (bits != 0) {
 		maskaddr.s_addr = htonl(0xFFFFFFFF << (32 - bits));
 		return &maskaddr;
@@ -636,6 +700,35 @@ parse_mask(char *mask)
 	}
 }
 
+void
+parse_direction(char *s)
+{
+	unsigned short direction;
+
+	if (strncmp("in", s, strlen(s)) == 0)
+#ifndef	IPFWADM_NO_ACCTDIR
+		direction = IP_FW_F_ACCTIN;
+#else	/* IPFWADM_NO_ACCTDIR */
+		exit_error(2, "accounting direction \"in\" not supported");
+#endif	/* IPFWADM_NO_ACCTDIR */
+	else if (strncmp("out", s, strlen(s)) == 0)
+#ifndef	IPFWADM_NO_ACCTDIR
+		direction = IP_FW_F_ACCTOUT;
+#else	/* IPFWADM_NO_ACCTDIR */
+		exit_error(2, "accounting direction \"out\" not supported");
+#endif	/* IPFWADM_NO_ACCTDIR */
+	else if (strncmp("both", s, strlen(s)) == 0)
+		direction = 0;
+	else {
+		fprintf(stderr, "%s: invalid direction \"%s\" specified\n", program, s);
+		exit_tryhelp(2);
+		/* make the compiler happy... */
+		return;
+	}
+	firewall.fw_flg |= direction;
+	return;
+}
+
 void
 parse_policy(char *s)
 {
@@ -645,13 +738,17 @@ parse_policy(char *s)
 		policy = IP_FW_F_ACCEPT;
 	else if (strncmp("deny", s, strlen(s)) == 0)
 		policy = 0; /* as opposed to IP_FW_F_ACCEPT */
-	else if (strncmp("masquerade", s, strlen(s)) == 0)
-		policy = IP_FW_F_ACCEPT | IP_FW_F_MASQ;
 	else if (strncmp("reject", s, strlen(s)) == 0)
 		policy = IP_FW_F_ICMPRPL;
-	else {
+	else if (strncmp("masquerade", s, strlen(s)) == 0) {
+		/* for backwards compatibility, better use "accept" and "-m" */
+		check_option(OPT_MASQUERADE, 'm');
+		policy = IP_FW_F_ACCEPT | IP_FW_F_MASQ;
+	} else {
 		fprintf(stderr, "%s: invalid policy \"%s\" specified\n", program, s);
-		exit(2);
+		exit_tryhelp(2);
+		/* make the compiler happy... */
+		return;
 	}
 	firewall.fw_flg |= policy;
 	return;
@@ -672,7 +769,9 @@ parse_protocol(char *s)
 		protocol = IP_FW_F_ICMP;
 	else {
 		fprintf(stderr, "%s: invalid protocol \"%s\" specified\n", program, s);
-		exit(2);
+		exit_tryhelp(2);
+		/* make the compiler happy... */
+		return;
 	}
 	firewall.fw_flg |= protocol;
 }
@@ -700,7 +799,7 @@ parse_all_ports(char **ports, unsigned short *portnumbers,
 			if (portnumbers[0] > portnumbers[1]) {
 				fprintf(stderr, "%s: invalid range of ports (%u > %u)\n",
 					program, portnumbers[0], portnumbers[1]);
-				exit(2);
+				exit_tryhelp(2);
 			}
 		}
 	}
@@ -717,13 +816,17 @@ parse_port(char *port, unsigned short kind)
 		/* ICMP types (given as port numbers) must be numeric! */
 		fprintf(stderr, "%s: invalid ICMP type \"%s\" specified\n",
 			program, port);
-		exit(2);
+		exit_tryhelp(2);
+		/* make the compiler happy... */
+		return 0;
 	} else if ((portnum = service_to_port(port, kind)) != -1)
 		return (unsigned short) portnum;
 	else {
 		fprintf(stderr, "%s: invalid port/service \"%s\" specified\n",
 			program, port);
-		exit(2);
+		exit_tryhelp(2);
+		/* make the compiler happy... */
+		return 0;
 	}
 }
 
@@ -743,7 +846,7 @@ store_port(char *port, unsigned short *nports, int rangeflag, char *ports[])
 	if (firewall.fw_nsp + firewall.fw_ndp > IP_FW_MAX_PORTS) {
 		fprintf(stderr, "%s: too many ports specified (maximum %d)\n",
 			program, IP_FW_MAX_PORTS);
-		exit(2);
+		exit_tryhelp(2);
 	}
 	ports[*nports - 1] = port;
 	return;
@@ -763,12 +866,12 @@ parse_hexbyte(char *s, unsigned char *b)
 		else {
 			fprintf(stderr, "%s: invalid hexbyte \"%s\" specified\n",
 				program, s);
-			exit(2);
+			exit_tryhelp(2);
 		}
 	} else {
 		fprintf(stderr, "%s: invalid hexbyte \"%s\" specified\n",
 			program, s);
-		exit(2);
+		exit_tryhelp(2);
 	}
 }
 
@@ -971,7 +1074,7 @@ policy_to_string(int policy)
 	case IP_FW_F_ACCEPT:
 		return "accept";
 	case IP_FW_F_ACCEPT | IP_FW_F_MASQ:
-		return "masquerade";
+		return "accept/masquerade";
 	case IP_FW_F_ICMPRPL:
 		return "reject";
 	default:
@@ -989,9 +1092,8 @@ add_delete_entries(int cmd, int chain)
 		for (j = 0; j < ndaddrs; j++) {
 			firewall.fw_dst.s_addr = daddrs[j].s_addr;
 			if (options & OPT_VERBOSE)
-				print_firewall(stdout, &firewall, FMT_NOCOUNTS | FMT_OPTIONS | FMT_TOS |
-					((chain == CHN_IN || chain == CHN_OUT|| chain == CHN_FWD) ?
-					0 : FMT_NOKIND) | FMT_VIA | FMT_NUMERIC | FMT_NOTABLE);
+				print_firewall(stdout, &firewall, FMT_NOCOUNTS | FMT_OPTIONS |
+					FMT_TOS | FMT_VIA | FMT_NUMERIC | FMT_NOTABLE);
 			ret |= do_setsockopt(cmd, &firewall, sizeof(firewall));
 		}
 	}
@@ -1060,11 +1162,9 @@ list_entries(int chain, char *mode)
 	ntotal += nread;
 
 	format = 0;
-	if (chain == CHN_IN || chain == CHN_OUT || chain == CHN_FWD) {
+	if (chain == CHN_IN || chain == CHN_OUT || chain == CHN_FWD)
 		if (!(options & OPT_EXTENDED))
 			format |= FMT_NOCOUNTS;
-	} else
-		format |= FMT_NOKIND;
 
 	if (options & OPT_NUMERIC)
 		format |= FMT_NUMERIC;
@@ -1158,8 +1258,12 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 				fprintf(fp, FMT("%10s ","%s "), "bytes");
 			}
 		}
-		if (!(format & FMT_NOKIND))
-			fputs("typ ", fp);
+		if (!(format & FMT_NOKIND)) {
+			if (chain == CHN_ACCT)
+				fprintf(fp, FMT("%-3s ","%s "), "dir");
+			else
+				fprintf(fp, FMT("%-5s ","%s "), "type");
+		}
 		fputs("prot ", fp);
 		if (format & FMT_OPTIONS)
 			fputs("opt  ", fp);
@@ -1204,14 +1308,25 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 	}
 
 	if (!(format & FMT_NOKIND)) {
-		if (flags & IP_FW_F_MASQ)
-			fprintf(fp, "msq ");
-		else if (flags & IP_FW_F_ACCEPT)
-			fprintf(fp, "acc ");
-		else if (flags & IP_FW_F_ICMPRPL)
-			fprintf(fp, "rej ");
-		else
-			fprintf(fp, "den ");
+		if (chain == CHN_ACCT) {
+			if (flags & IP_FW_F_ACCTIN)
+				fprintf(fp, FMT("%-3s ", "%s "), "in");
+			else if (flags & IP_FW_F_ACCTOUT)
+				fprintf(fp, FMT("%-3s ", "%s "), "out");
+			else
+				fprintf(fp, FMT("%-3s ", "%s "), "i/o");
+		} else {
+			if (flags & IP_FW_F_REDIR)
+				fprintf(fp, FMT("%-5s ", "%s "), "acc/r");
+			else if (flags & IP_FW_F_MASQ)
+				fprintf(fp, FMT("%-5s ", "%s "), "acc/m");
+			else if (flags & IP_FW_F_ACCEPT)
+				fprintf(fp, FMT("%-5s ", "%s "), "acc");
+			else if (flags & IP_FW_F_ICMPRPL)
+				fprintf(fp, FMT("%-5s ", "%s "), "rej");
+			else
+				fprintf(fp, FMT("%-5s ", "%s "), "deny");
+		}
 	}
 
 	switch (kind) {
@@ -1328,6 +1443,16 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 				fprintf(fp, "%u", fw->fw_pts[i]);
 		}
 
+	if (flags & IP_FW_F_REDIR) {
+		i = fw->fw_nsp + fw->fw_ndp;
+		if (format & FMT_NUMERIC)
+			fprintf(fp, " => %u", fw->fw_pts[i]);
+		else if ((service = port_to_service(fw->fw_pts[i], kind)) != NULL)
+			fprintf(fp, " => %s", service);
+		else
+			fprintf(fp, " => %u", fw->fw_pts[i]);
+	}
+
 	if (!(format & FMT_NONEWLINE))
 		putc('\n', fp);
 }
@@ -1378,16 +1503,16 @@ print_masq(FILE *fp, struct masq *ms, int format)
 	}
 
 	if (format & FMT_NUMERIC)
-		fprintf(fp, "%u (%u) -> %u", ms->sport, ms->mport, ms->dport);
+		fprintf(fp, "%u (%u) -> %u\n", ms->sport, ms->mport, ms->dport);
 	else {
 		if ((service = port_to_service(ms->sport, ms->kind)) != NULL)
 			fprintf(fp, "%s (%u) -> ", service, ms->mport);
 		else
 			fprintf(fp, "%u (%u) -> ", ms->sport, ms->mport);
 		if ((service = port_to_service(ms->dport, ms->kind)) != NULL)
-			fputs(service, fp);
+			fprintf(fp, "%s\n", service);
 		else
-			fprintf(fp, "%u", ms->dport);
+			fprintf(fp, "%u\n", ms->dport);
 	}
 }
 
@@ -1546,12 +1671,12 @@ do_setsockopt(int cmd, void *data, int length)
 }
 
 void
-check_option(int option, char name)
+check_option(long option, char name)
 {
 	if (options & option) {
 		fprintf(stderr, "%s: multiple -%c flags not allowed\n",
 			program, name);
-		exit(2);
+		exit_tryhelp(2);
 	}
 	options |= option;
 }
@@ -1617,7 +1742,7 @@ void
 exit_printhelp()
 {
 	printf("%s\n\n"
-		"Usage: %s -A command [options] (accounting)\n"
+		"Usage: %s -A [direction] command [options] (accounting)\n"
 		"       %s -F command [options] (forwarding firewall)\n"
 		"       %s -I command [options] (input firewall)\n"
 		"       %s -O command [options] (output firewall)\n"
@@ -1633,7 +1758,7 @@ exit_printhelp()
 		"  -l		list all rules of this category\n"
 		"  -z		reset packet/byte counters of all rules of this category\n"
 		"  -f		remove all rules of this category\n"
-		"  -p policy	change default policy (accept/masquerade/deny/reject)\n"
+		"  -p policy	change default policy (accept/deny/reject)\n"
 		"  -c		check acceptance of IP packet\n\n"
 		"Options:\n"
 		"  -P		protocol (either tcp, udp, icmp, or all)\n"
@@ -1641,17 +1766,19 @@ exit_printhelp()
 		"		source specification\n"
 		"  -D address[/mask] [port ...]\n"
 		"		destination specification\n"
-		"  -V address	interface address\n"
-		"  -W name	interface name\n"
+		"  -V address	network interface address\n"
+		"  -W name	network interface name\n"
 		"  -b		bidirectional match\n"
 		"  -e		extended output mode\n"
-		"  -k		match TCP packets with ACK set\n"
+		"  -k		match TCP packets only when ACK set\n"
+		"  -m		masquerade packets as coming from local host\n"
 		"  -n		numeric output of addresses and ports\n"
 		"  -o		turn on kernel logging for matching packets\n"
+		"  -r port	redirect packets to local port (transparent proxying)\n"
 		"  -t and xor	and/xor masks for TOS field\n"
 		"  -v		verbose mode\n"
 		"  -x		expand numbers (display exact values)\n"
-		"  -y		match TCP packets with SYN set and ACK cleared\n");
+		"  -y		match TCP packets only when SYN set and ACK cleared\n");
 
 	exit(0);
 }
